-- THIS SCRIPT WAS MADE BY SOLARPOWER07
-- DO NOT COPY OR REDISTRIBUTE WITHOUT THE ABOVE MESSAGE!

local M = {}

local _internalCoef = 0.65
-- local _internalDragCoef = 0
local _internalDragCoef = 0.5
local _internalSideForceCoef = 0.05
local _internalRoofFlapCoef = 0.5
local _internalGroundOppositeEffect = 2
local _internalInvertedCurveOffset = -0.25
local blowoverCoef = 1
local dragCoef = 1
local roofFlapCoef = 1
local sideForceCoef = 1
local sideCoefOffset = 1
local curveOffset = 0
local curveOffsetFront = 0
local invertedCurveOffset = 0
local blowoversEnabled = false
local __DEBUG_PRINT_HEIGHT = false
local __DEBUG_PRINT = false

-- Specific pressure component coefficients
local groundCoef = 1
local groundOppositeEffect = 1
local groundLowPressureCoef = 1

-- Template for the models

local frontModel = {
	nodes = {},
	count = 0,
	refHeight = 0,
	floorRoughness = 0,
	curveOffset = 0,
	liftCoef = 0,
	dragCoef = 0,
	roofFlap = 0
}

local rearModel = {
	nodes = {},
	count = 0,
	refHeight = 0,
	floorRoughness = 0,
	curveOffset = 0,
	liftCoef = 0,
	dragCoef = 0,
	roofFlap = 0
}

local leftModel = {
	nodes = {},
	count = 0,
	refHeight = 0,
	floorRoughness = 0,
	curveOffset = 0,
	liftCoef = 0,
	dragCoef = 0,
	roofFlap = 0
}

local rightModel = {
	nodes = {},
	count = 0,
	refHeight = 0,
	floorRoughness = 0,
	curveOffset = 0,
	liftCoef = 0,
	dragCoef = 0,
	roofFlap = 0
}

local globalUp = vec3(0,0,1)

-- Helper functions

-- Calculates the height of a vector off of the ground
local function calcHeight(vec)
	-- Casts a ray from the vector downwards to the ground, and returns the distance
	return obj:castRayStatic(vec, vec3(0, 0, -1), 1000)
end

local function sign(val)
	if val > 0 then return 1 end
	if val < 0 then return -1 end
	return 0
end

local function square(val)
	return val * val
end

local function sigmoid(x)
	return 1 / (1 + math.exp(-x))
end

local function clamp(val)
	return math.max(math.min(val, 1), 0)
end

-- Calculates the air pressure on a side of the car
function calculateAirPressure(sideData, oppositeData, carData, density)

    -- The higher the roughness, the more the air can dam under the car instead of being compressed and increasing the pressure.
	-- Thus, this lowers the blowover force a bit close to the ground.
	-- Note: floor roughness is used as a coefficient for other things, such as side force and ground opposite coef!
    local roughnessEffect = clamp(1 - sideData.roughness * (1 - math.abs(carData.vectors.bottom:dot(carData.vel) / carData.speed)))

	-- Limits blowover forces when on the ground and facing the direction of the side. Is overpowered once there is enough of a gap,
	--  and depending on the situation, only slows down the start of blowovers instead of stopping them entirely
	local roofFlapEffect = clamp(1 - sideData.roofFlap * sigmoid(-sideData.vector:dot(carData.vel) - 20) * roofFlapCoef * _internalRoofFlapCoef)

    -- Limits blowover forces when extremely close to the ground (as air cannot get in unless theres enough of a gap)
	--  and slowly tapers off blowover forces the higher off the ground the car goes (air is no longer trapped between the car and the ground)
    local groundEffect = clamp(carData.coefficients.ground *
        (1 - math.min((math.min(sideData.height, oppositeData.height) / (sideData.wingSpan * 2)) ^ 2, 1)) *
		sigmoid((((sideData.height * 25) - 0.5) / carData.coefficients.groundLowPressure) - 0))
	
	-- Limits the forces when the car is inverted
	local invertedEffect = clamp(1 - ((1 - (math.max(0, carData.vectors.up.z + (_internalInvertedCurveOffset + invertedCurveOffset)) / (1 + (_internalInvertedCurveOffset + invertedCurveOffset)))) ^ 2))

    -- Main coef, determines the base lifting force from the angle of attack
    local mainCoef = math.max(-sideData.vector:dot(carData.vel) / carData.speed, 0) *
        math.max(((carData.vectors.bottom:dot(carData.vel) / carData.speed) + sideData.curveOffset) / (1 + sideData.curveOffset), 0)

	-- Roll coef, prevents blowover forces from slowing down a roll that is ongoing
	local rollCoef = math.max(1 - math.max(sideData.rollCoef + 0.5, 0) / 3, 0)

    return (0.5 * density * carData.speed ^ 2) * mainCoef * _internalCoef * roughnessEffect * groundEffect * invertedEffect * roofFlapEffect * rollCoef * blowoverCoef
end

-- Side force calculation
local function sideForce(direction, up, velocity, yawSpeed, area)
	local sideVector = direction:cross(up)
	local dirCoef = direction:dot(velocity) / velocity:length()
	local sideCoef = sideVector:dot(velocity) / velocity:length() * sign(yawSpeed)
	local yawSpeedWithOffset = yawSpeed + (-sign(sideCoef) * 2 * sideCoefOffset)

	local sideForceMul = blowoverCoef
	if sideForceMul > 2 then
		sideForceMul = 2
	end

	return math.max((dirCoef+1) * (sideCoef+0.5), 0) * yawSpeedWithOffset * square(velocity:length()) * area * _internalSideForceCoef * _internalCoef * sideForceMul * sideForceCoef
end

-- Cache the forces
local frontForce = vec3(0,0,0)
local rearForce = vec3(0,0,0)
local leftForce = vec3(0,0,0)
local rightForce = vec3(0,0,0)

local dragForce = vec3(0,0,0)

local frontSideForce = vec3(0,0,0)
local rearSideForce = vec3(0,0,0)
local leftSideForce = vec3(0,0,0)
local rightSideForce = vec3(0,0,0)

local function updateForces()

	-- Gets the object's direction vectors
	local forward = -obj:getDirectionVector()
	local up = obj:getDirectionVectorUp()
	local left = forward:cross(up)
	local right = up:cross(forward)
	local back = left:cross(up)
	local down = forward:cross(left)

	-- Get the velocity vector, subtracting wind speed
	local velocity = obj:getVelocity()
	local windVector = obj:getFlow()
	velocity = velocity - windVector

	local speed = velocity:length()

	-- Return if speed is NaN or 0
	if speed ~= speed or speed == 0 then return end

	if __DEBUG_PRINT then
		log("I","mods.blowover_mod_v2","New Tick:")
	end

	-- Get the velocity vector, relative to the car
	local velocityVector = vec3(velocity:dot(left), velocity:dot(back), velocity:dot(up))

	-- Get the area of the floor that is facing the direction of the velocity
	local floorArea = obj:getInitialLength() * obj:getInitialWidth()
	local exposedArea = math.max(floorArea * down:dot(velocity) / speed, 0)
	
	-- Wingspan of the vehicle, tells it when to stop applying the blowover forces
	local wingSpanLength = obj:getInitialLength()
	local wingSpanWidth = obj:getInitialWidth()

	-- Get heights of the reference points off of the ground, to allow ground clearance in air pressure calculations

	-- Starts off as the average global position of the reference points
	local frontRefHeight = vec3(0,0,0)
	local rearRefHeight = vec3(0,0,0)
	local leftRefHeight = vec3(0,0,0)
	local rightRefHeight = vec3(0,0,0)

	for i=0,3 do
		if frontModel.nodes[i] ~= nil then
			frontRefHeight = frontRefHeight + obj:getNodePosition(frontModel.nodes[i])
		end
		if rearModel.nodes[i] ~= nil then
			rearRefHeight = rearRefHeight + obj:getNodePosition(rearModel.nodes[i])
		end
		if leftModel.nodes[i] ~= nil then
			leftRefHeight = leftRefHeight + obj:getNodePosition(leftModel.nodes[i])
		end
		if rightModel.nodes[i] ~= nil then
			rightRefHeight = rightRefHeight + obj:getNodePosition(rightModel.nodes[i])
		end
	end
	
	frontRefHeight = math.max(calcHeight((frontRefHeight / frontModel.count) + obj:getPosition()) - frontModel.refHeight, 0)
	rearRefHeight = math.max(calcHeight((rearRefHeight / rearModel.count) + obj:getPosition()) - rearModel.refHeight, 0)
	leftRefHeight = math.max(calcHeight((leftRefHeight / leftModel.count) + obj:getPosition()) - leftModel.refHeight, 0)
	rightRefHeight = math.max(calcHeight((rightRefHeight / rightModel.count) + obj:getPosition()) - rightModel.refHeight, 0)

	-- Print for debugging
	if __DEBUG_PRINT_HEIGHT then
		log("D","mods.blowover_mod_v2","Front height: "..frontRefHeight)
		log("D","mods.blowover_mod_v2","Rear height: "..rearRefHeight)
		log("D","mods.blowover_mod_v2","Left height: "..leftRefHeight)
		log("D","mods.blowover_mod_v2","Right height: "..rightRefHeight)
	end

	local carData = {
		vel = velocity,
		speed = speed,
		vectors = {
			front = forward,
			back = back,
			left = left,
			right = right,
			up = up,
			down = down,
			bottom = down
		},
		coefficients = {
			ground = groundCoef,
			groundLowPressure = groundLowPressureCoef
		},
		length = obj:getInitialLength(),
		width = obj:getInitialWidth()
	}

	local flipped = up.z < 0

	local rollAngularVelocity = obj:getRollAngularVelocity()
	local pitchAngularVelocity = obj:getPitchAngularVelocity()

	local _sideveloc = left:dot(velocity)
	if flipped then _sideveloc = right:dot(velocity) end
	local _sidevelocsign = 0
	if _sideveloc ~= 0 then _sidevelocsign = _sideveloc / math.abs(_sideveloc) end

	local _lengthveloc = forward:dot(velocity)
	if flipped then _lengthveloc = back:dot(velocity) end
	local _lengthvelocsign = 0
	if _lengthveloc ~= 0 then _lengthvelocsign = _lengthveloc / math.abs(_lengthveloc) end

	-- Data for each side
	local frontData = {
		vector = forward,
		height = frontRefHeight,
		roughness = frontModel.floorRoughness,
		liftCoef = frontModel.liftCoef,
		roofFlap = frontModel.roofFlap,
		curveOffset = frontModel.curveOffset + curveOffsetFront,
		wingSpan = wingSpanLength,
		rollCoef = pitchAngularVelocity * _lengthvelocsign,
		__name = "front"
	}
	local rearData = {
		vector = back,
		height = rearRefHeight,
		roughness = rearModel.floorRoughness,
		liftCoef = rearModel.liftCoef,
		roofFlap = rearModel.roofFlap,
		curveOffset = rearModel.curveOffset + curveOffset,
		wingSpan = wingSpanLength,
		rollCoef = pitchAngularVelocity * -_lengthvelocsign,
		__name = "rear"
	}
	local leftData = {
		vector = right,
		height = leftRefHeight,
		roughness = leftModel.floorRoughness,
		liftCoef = leftModel.liftCoef,
		roofFlap = leftModel.roofFlap,
		curveOffset = leftModel.curveOffset + curveOffset,
		wingSpan = wingSpanWidth,
		rollCoef = rollAngularVelocity * _sidevelocsign,
		__name = "left"
	}
	local rightData = {
		vector = left,
		height = rightRefHeight,
		roughness = rightModel.floorRoughness,
		liftCoef = rightModel.liftCoef,
		roofFlap = rightModel.roofFlap,
		curveOffset = rightModel.curveOffset + curveOffset,
		wingSpan = wingSpanWidth,
		rollCoef = rollAngularVelocity * -_sidevelocsign,
		__name = "right"
	}

	-- Calculate the air pressures on each side
	local frontStaticPressure = calculateAirPressure(frontData, rearData, carData, obj:getAirDensity())
	local rearStaticPressure = calculateAirPressure(rearData, frontData, carData, obj:getAirDensity())
	local leftStaticPressure = calculateAirPressure(leftData, rightData, carData, obj:getAirDensity())
	local rightStaticPressure = calculateAirPressure(rightData, leftData, carData, obj:getAirDensity())

	-- Remove NaN values
	if frontStaticPressure ~= frontStaticPressure then frontStaticPressure = 0 end
	if rearStaticPressure ~= rearStaticPressure then rearStaticPressure = 0 end
	if leftStaticPressure ~= leftStaticPressure then leftStaticPressure = 0 end
	if rightStaticPressure ~= rightStaticPressure then rightStaticPressure = 0 end

	-- Print for debugging
	if __DEBUG_PRINT then
		log("D","mods.blowover_mod_v2","Front static pressure: "..frontStaticPressure)
		log("D","mods.blowover_mod_v2","Rear static pressure: "..rearStaticPressure)
		log("D","mods.blowover_mod_v2","Left static pressure: "..leftStaticPressure)
		log("D","mods.blowover_mod_v2","Right static pressure: "..rightStaticPressure)
	end

	-- Account for the opposite side's pressure
	local frontPressure = frontStaticPressure + rearStaticPressure * groundOppositeEffect * _internalGroundOppositeEffect * frontData.roughness
	local rearPressure = rearStaticPressure + frontStaticPressure * groundOppositeEffect * _internalGroundOppositeEffect * rearData.roughness
	local leftPressure = leftStaticPressure + rightStaticPressure * groundOppositeEffect * _internalGroundOppositeEffect * leftData.roughness
	local rightPressure = rightStaticPressure + leftStaticPressure * groundOppositeEffect * _internalGroundOppositeEffect * rightData.roughness

	-- Print for debugging
	if __DEBUG_PRINT then
		log("D","mods.blowover_mod_v2","Front pressure: "..frontPressure)
		log("D","mods.blowover_mod_v2","Rear pressure: "..rearPressure)
		log("D","mods.blowover_mod_v2","Left pressure: "..leftPressure)
		log("D","mods.blowover_mod_v2","Right pressure: "..rightPressure)
	end

	-- Calculate the total lift force
	local frontLift = frontPressure * frontData.liftCoef * floorArea
	local rearLift = rearPressure * rearData.liftCoef * floorArea
	local leftLift = leftPressure * leftData.liftCoef * floorArea
	local rightLift = rightPressure * rightData.liftCoef * floorArea

	-- Print for debugging
	if __DEBUG_PRINT then
		log("D","mods.blowover_mod_v2","Front lift: "..frontLift)
		log("D","mods.blowover_mod_v2","Rear lift: "..rearLift)
		log("D","mods.blowover_mod_v2","Left lift: "..leftLift)
		log("D","mods.blowover_mod_v2","Right lift: "..rightLift)
	end
	
    -- With the lift, find the force on each end of the floor, in vector form
    -- The force is perpendicular to the floor, so it is in the direction of the up vector of the car
	frontForce = frontLift * up
	rearForce = rearLift * up
	leftForce = leftLift * up
	rightForce = rightLift * up

	-- Print for debugging
	if __DEBUG_PRINT then
		log("D","mods.blowover_mod_v2","Front force: "..frontForce:length())
		log("D","mods.blowover_mod_v2","Rear force: "..rearForce:length())
		log("D","mods.blowover_mod_v2","Left force: "..leftForce:length())
		log("D","mods.blowover_mod_v2","Right force: "..rightForce:length())
	end

	-- Find drag force, proportional to the square of the speed
	local exposedProportion = clamp(exposedArea / floorArea)
	dragForce = -velocity * 0.5 * speed * obj:getAirDensity() * dragCoef * exposedProportion * _internalDragCoef * _internalCoef

	-- Print for debugging (drag force magnitude)
	if __DEBUG_PRINT then
		log("D","mods.blowover_mod_v2","Drag force: "..tostring(dragForce:length()))
	end

	-- Side force, proportional to the square of the speed
	local exposedFloorCurve = exposedProportion * ((1 - exposedProportion) ^ 2) * 4
	local yawSpeed = obj:getYawAngularVelocity()

	frontSideForce = sideForce(-forward, up, velocity, yawSpeed, exposedFloorCurve * floorArea) * left * frontData.roughness
	rearSideForce = sideForce(-back, up, velocity, yawSpeed, exposedFloorCurve * floorArea) * right * rearData.roughness
	leftSideForce = sideForce(-left, up, velocity, yawSpeed, exposedFloorCurve * floorArea) * forward * leftData.roughness
	rightSideForce = sideForce(-right, up, velocity, yawSpeed, exposedFloorCurve * floorArea) * back * rightData.roughness
end

-- Current time
local currentTime = 0
local forcesSet = false

local updateInterval = 0.01

local function reset()
	currentTime = 0
	forcesSet = false
end

local function update(deltaTime)

	if not blowoversEnabled then return end

	-- Update the forces on the car
	if not forcesSet then
		updateForces()
		forcesSet = true
	else
		-- Update the forces every updateInterval seconds
		currentTime = currentTime + deltaTime
		if currentTime >= updateInterval then
			currentTime = 0
			updateForces()
		end
	end

	-- Apply the forces to the car
	for i=0,3 do
		if frontModel.nodes[i] ~= nil then -- * obj:getNodeMass(frontModel.nodes[i])
			obj:applyForceVector(frontModel.nodes[i], frontForce / frontModel.count)
			obj:applyForceVector(frontModel.nodes[i], dragForce * frontModel.dragCoef / frontModel.count)
			obj:applyForceVector(frontModel.nodes[i], frontSideForce / frontModel.count)
		end if rearModel.nodes[i] ~= nil then
			obj:applyForceVector(rearModel.nodes[i], rearForce / rearModel.count)
			obj:applyForceVector(rearModel.nodes[i], dragForce * rearModel.dragCoef / rearModel.count)
			obj:applyForceVector(rearModel.nodes[i], rearSideForce / rearModel.count)
		end if leftModel.nodes[i] ~= nil then
			obj:applyForceVector(leftModel.nodes[i], leftForce / leftModel.count)
			obj:applyForceVector(leftModel.nodes[i], dragForce * leftModel.dragCoef / leftModel.count)
			obj:applyForceVector(leftModel.nodes[i], leftSideForce / leftModel.count)
		end if rightModel.nodes[i] ~= nil then
			obj:applyForceVector(rightModel.nodes[i], rightForce / rightModel.count)
			obj:applyForceVector(rightModel.nodes[i], dragForce * rightModel.dragCoef / rightModel.count)
			obj:applyForceVector(rightModel.nodes[i], rightSideForce / rightModel.count)
		end
	end
end

local function init()
	log("I","mods.blowover_mod_v2","STARTING BLOWOVER MOD (by Solarpower07)")

	if v.data.blowoverModel ~= nil then
		blowoverCoef = v.data.blowoverModel.blowoverCoefficient or 1
		dragCoef = v.data.blowoverModel.dragCoefficient or 1
		roofFlapCoef = v.data.blowoverModel.roofFlapCoefficient or 1
		sideForceCoef = v.data.blowoverModel.sideForceCoefficient or 1
		sideCoefOffset = v.data.blowoverModel.sideCoefficientOffset or 1
		curveOffset = v.data.blowoverModel.curveOffset or 0
		curveOffsetFront = v.data.blowoverModel.curveOffsetFront or 0
		if curveOffset < 0 then curveOffset = 0 end

		groundCoef = v.data.blowoverModel.groundCoefficient or 1
		groundOppositeEffect = v.data.blowoverModel.groundOppositeEffect or 1
		groundLowPressureCoef = v.data.blowoverModel.groundLowPressureCoefficient or 1
		__DEBUG_PRINT_HEIGHT = v.data.blowoverModel.DEBUG_PRINT_HEIGHT or false
		__DEBUG_PRINT = v.data.blowoverModel.DEBUG_PRINT or false
		log("I","mods.blowover_mod_v2","Global blowover coefficients loaded")

		-- Print the model to test if it's loaded correctly
		if __DEBUG_PRINT then
			dump(v.data.blowoverModel)
		end
	end

	-- Load values from the models
	if v.data.frontBlowoverModel ~= nil then
		frontModel.refHeight = v.data.frontBlowoverModel.refHeight or 0
		frontModel.floorRoughness = v.data.frontBlowoverModel.floorRoughness or 0
		frontModel.curveOffset = v.data.frontBlowoverModel.curveOffset or 0
		if frontModel.curveOffset < 0 then frontModel.curveOffset = 0 end
		frontModel.liftCoef = v.data.frontBlowoverModel.liftCoef or 0
		frontModel.dragCoef = v.data.frontBlowoverModel.dragCoef or 0
		frontModel.roofFlap = v.data.frontBlowoverModel.roofFlap or 0
		if v.data.blowoverModel.roofFlaps ~= nil and not v.data.blowoverModel.roofFlaps then frontModel.roofFlap = 0 end
	end
	if v.data.rearBlowoverModel ~= nil then
		rearModel.refHeight = v.data.rearBlowoverModel.refHeight or 0
		rearModel.floorRoughness = v.data.rearBlowoverModel.floorRoughness or 0
		rearModel.curveOffset = v.data.rearBlowoverModel.curveOffset or 0
		if rearModel.curveOffset < 0 then rearModel.curveOffset = 0 end
		rearModel.liftCoef = v.data.rearBlowoverModel.liftCoef or 0
		rearModel.dragCoef = v.data.rearBlowoverModel.dragCoef or 0
		rearModel.roofFlap = v.data.rearBlowoverModel.roofFlap or 0
		if v.data.blowoverModel.roofFlaps ~= nil and not v.data.blowoverModel.roofFlaps then rearModel.roofFlap = 0 end
	end
	if v.data.leftBlowoverModel ~= nil then
		leftModel.refHeight = v.data.leftBlowoverModel.refHeight or 0
		leftModel.floorRoughness = v.data.leftBlowoverModel.floorRoughness or 0
		leftModel.curveOffset = v.data.leftBlowoverModel.curveOffset or 0
		if leftModel.curveOffset < 0 then leftModel.curveOffset = 0 end
		leftModel.liftCoef = v.data.leftBlowoverModel.liftCoef or 0
		leftModel.dragCoef = v.data.leftBlowoverModel.dragCoef or 0
		leftModel.roofFlap = v.data.leftBlowoverModel.roofFlap or 0
		if v.data.blowoverModel.roofFlaps ~= nil and not v.data.blowoverModel.roofFlaps then leftModel.roofFlap = 0 end
	end
	if v.data.rightBlowoverModel ~= nil then
		rightModel.refHeight = v.data.rightBlowoverModel.refHeight or 0
		rightModel.floorRoughness = v.data.rightBlowoverModel.floorRoughness or 0
		rightModel.curveOffset = v.data.rightBlowoverModel.curveOffset or 0
		if rightModel.curveOffset < 0 then rightModel.curveOffset = 0 end
		rightModel.liftCoef = v.data.rightBlowoverModel.liftCoef or 0
		rightModel.dragCoef = v.data.rightBlowoverModel.dragCoef or 0
		rightModel.roofFlap = v.data.rightBlowoverModel.roofFlap or 0
		if v.data.blowoverModel.roofFlaps ~= nil and not v.data.blowoverModel.roofFlaps then rightModel.roofFlap = 0 end
	end
	if __DEBUG_PRINT then
		log("D","mods.blowover_mod_v2","Front model:")
		dump(frontModel)
		log("D","mods.blowover_mod_v2","Rear model:")
		dump(rearModel)
		log("D","mods.blowover_mod_v2","Left model:")
		dump(leftModel)
		log("D","mods.blowover_mod_v2","Right model:")
		dump(rightModel)
	end
	log("I","mods.blowover_mod_v2","Blowover models loaded")

	if v.data.blowoverRefs ~= nil then
		for k,v in pairs(v.data.blowoverRefs) do
			if k == 0 then
				frontModel.nodes[0] = v.ref1
				frontModel.nodes[1] = v.ref2
				frontModel.nodes[2] = v.ref3
				frontModel.nodes[3] = v.ref4
				if __DEBUG_PRINT then
					log("D","mods.blowover_mod_v2","Front nodes: "..tostring(frontModel.nodes[0])..", "..tostring(frontModel.nodes[1])..", "..tostring(frontModel.nodes[2])..", "..tostring(frontModel.nodes[3]))
				end
			elseif k == 1 then
				rearModel.nodes[0] = v.ref1
				rearModel.nodes[1] = v.ref2
				rearModel.nodes[2] = v.ref3
				rearModel.nodes[3] = v.ref4
				if __DEBUG_PRINT then
					log("D","mods.blowover_mod_v2","Rear nodes: "..tostring(rearModel.nodes[0])..", "..tostring(rearModel.nodes[1])..", "..tostring(rearModel.nodes[2])..", "..tostring(rearModel.nodes[3]))
				end
			elseif k == 2 then
				leftModel.nodes[0] = v.ref1
				leftModel.nodes[1] = v.ref2
				leftModel.nodes[2] = v.ref3
				leftModel.nodes[3] = v.ref4
				if __DEBUG_PRINT then
					log("D","mods.blowover_mod_v2","Left nodes: "..tostring(leftModel.nodes[0])..", "..tostring(leftModel.nodes[1])..", "..tostring(leftModel.nodes[2])..", "..tostring(leftModel.nodes[3]))
				end
			elseif k == 3 then
				rightModel.nodes[0] = v.ref1
				rightModel.nodes[1] = v.ref2
				rightModel.nodes[2] = v.ref3
				rightModel.nodes[3] = v.ref4
				if __DEBUG_PRINT then
					log("D","mods.blowover_mod_v2","Right nodes: "..tostring(rightModel.nodes[0])..", "..tostring(rightModel.nodes[1])..", "..tostring(rightModel.nodes[2])..", "..tostring(rightModel.nodes[3]))
				end
			end
		end

		-- Get count
		for i=0,3 do
			if frontModel.nodes[i] ~= nil then
				frontModel.count = i+1
			end if rearModel.nodes[i] ~= nil then
				rearModel.count = i+1
			end if leftModel.nodes[i] ~= nil then
				leftModel.count = i+1
			end if rightModel.nodes[i] ~= nil then
				rightModel.count = i+1
			end
		end
		log("I","mods.blowover_mod_v2","Blowover reference nodes loaded")
	end

	blowoversEnabled = true

	log("I","mods.blowover_mod_v2","BLOWOVERS ENABLED, EVERYTHING OKAY")
end


-- public interface
M.reset = reset
M.init = init
M.update = update
M.updateGFX = nop

return M
